{-# OPTIONS -cpp #-}

module Scheduler ( Context
                 , initContext
                 , handleEvents, {-chandleEvents,-}
                 , ioProcess, handleContextOSEvent
                 , addInteractiveProcess
                 , closeContext
                 , quitProcess
                 ) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	Scheduler contains the process creation, termination, and handling functions.
--	********************************************************************************


import ClCrossCall_12 (osInitToolbox)	-- Should actually be in a Os... module, but not OSToolBox
import CommonDef
import Id
import IOState
import OSEvent
import StdProcessDef
import World
import TimerTable(getTimeIntervalFromTimerTable)
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


schedulerFatalError :: String -> String -> x
schedulerFatalError rule message
	= dumpFatalError rule "Scheduler" message


--	Access functions on RuntimeState:

rsIsClosed :: RuntimeState -> Bool
rsIsClosed Closed = True
rsIsClosed _      = False


--	Starting an interactive process.

initContext :: [ProcessAttribute IF_MVAR(,ps)] -> ProcessInit IF_MVAR(,ps -> ps) -> DocumentInterface -> IO (Maybe Context)
initContext pAtts initIO IF_MVAR(,ps) xDI
	= do {
		initOK <- osInitToolbox;							-- initialise toolbox
		if   initOK									-- toolbox correctly initialised
		then do {
			w <- loadWorld;
			storeWorld w;
			context  <- initialContext w Nothing;					-- initialise context
			context1 <- createInitialProcess pAtts initIO IF_MVAR(,ps) xDI context;	-- fork the initial interactive process
			return (Just context1)
		     }
		else return Nothing;		
	  }


--	Handling events until termination of all interactive processes.

handleEvents :: Context -> IO Context
handleEvents context
	= osHandleEvents
		terminate
		getosevents
		setosevents
		getsleeptime
		handleContextOSEvent
		context
	where
		terminate :: Context -> IO (Bool,Context)
		terminate context
			= do {
				iocontext <- takeMVar context;
				let (empty,iocontext1) = ioContextEmptyProcessEventHandlers iocontext				
				in putMVar context iocontext1 >> return (empty,context)
			  }
		
		getosevents :: Context -> IO (OSEvents,Context)
		getosevents context = do
			iocontext <- takeMVar context
			let (osevents,iocontext1) = ioContextGetEvents iocontext
			putMVar context iocontext1
			return (osevents,context)

		
		setosevents :: (OSEvents,Context) -> IO Context
		setosevents (osevents,context)
			= do {
				iocontext <- takeMVar context;
				putMVar context (ioContextSetEvents osevents iocontext) >> return context
			  }
			  
		getsleeptime :: Context -> IO (Int,Context)
		getsleeptime context = do
			iocontext <- takeMVar context
			let (tt, iocontext1) = ioContextGetTimerTable iocontext
			let maybe_sleep	= getTimeIntervalFromTimerTable tt
			let sleep = maybe osLongSleep id maybe_sleep
			putMVar context iocontext1
			return (sleep, context)

--	Closing a final context. 

closeContext :: Context -> IO ()
closeContext context
	= do {		
		iocontext <- readMVar context;		
		if   fst (ioContextEmptyProcessEventHandlers iocontext)
		then return ()
		else schedulerFatalError "closeContext" "not a final Context";		
	  }


handleContextOSEvent :: SchedulerEvent -> Context -> IO ([Int],Context)
handleContextOSEvent schedulerEvent context = do
	iocontext <- takeMVar context
	let (processfilters,iocontext1) = ioContextGetProcessEventHandlers iocontext
	putMVar context iocontext1
	schedulerEvent1 <- letProcessesFilterEvent processfilters schedulerEvent;
	let replyToOS = case schedulerEvent1 of
		(ScheduleOSEvent _ reply) -> reply
		_                         -> []
	return (replyToOS, context)
	where
		letProcessesFilterEvent :: [ProcessEventHandler] -> SchedulerEvent -> IO SchedulerEvent
		letProcessesFilterEvent [] 		schedulerEvent = return schedulerEvent
		letProcessesFilterEvent (peh : pehs) 	schedulerEvent
			= do {
				(ok,maybeDeviceEvent,schedulerEvent1) <- filterfun schedulerEvent;
				if   not ok
				then letProcessesFilterEvent pehs schedulerEvent1
				else if   isNothing maybeDeviceEvent
				     then return schedulerEvent1
				     else 
				     do {
				          writeDeviceEvent de (fromJust maybeDeviceEvent);
				          return schedulerEvent1
				     }
			  }
			where
				filterfun = pehEventFilter  peh
				de        = pehDeviceEvents peh		


createInitialProcess :: [ProcessAttribute IF_MVAR(,ps)] -> ProcessInit IF_MVAR(,ps -> ps) -> DocumentInterface -> Context -> IO Context
createInitialProcess pAtts initIO IF_MVAR(,ps) xDI context = do
    de        <- newDeviceEvents
    iocontext <- takeMVar context
    let (ioId,iocontext1) = ioContextNewMaxIONr iocontext
    newIOSt   <- emptyIOSt ioId xDI pAtts initIO context de;
    let processEventHandler = ProcessEventHandler
		{ pehId           = ioId
		, pehEventFilter  = processEventFilter newIOSt
		, pehDeviceEvents = de
		}
    let iocontext2 = ioContextSetProcessEventHandler processEventHandler iocontext1
    putMVar context iocontext2
    threadId <- forkIO (ioProcess de ps newIOSt >> return ())	-- Spawn the interactive process (it blocks immediately on channel)    
    writeDeviceEvent de (ProcessDevice,ProcessInitialise)	-- Make sure the interactive process gets started    
    return context

#if MVAR
addInteractiveProcess :: [ProcessAttribute] -> ProcessInit -> DocumentInterface -> GUI ()
addInteractiveProcess pAtts initIO xDI
#else
addInteractiveProcess :: [ProcessAttribute ps'] -> ProcessInit ps' -> ps' -> DocumentInterface -> GUI ps ()
addInteractiveProcess pAtts initIO ps' xDI
#endif
	= do {
		ioId      <- ioStNewMaxIONr;
		de        <- liftIO newDeviceEvents;
		context   <- accIOEnv ioStGetContext;
		newIOSt   <- liftIO (emptyIOSt ioId xDI pAtts initIO context de);
		iocontext <- liftIO (takeMVar context);
		let processEventHandler = ProcessEventHandler
		                              { pehId           = ioId
		                              , pehEventFilter  = processEventFilter newIOSt
		                              , pehDeviceEvents = de
		                              }
		    iocontext1          = ioContextSetProcessEventHandler processEventHandler iocontext
		in
		do {
			liftIO (putMVar context iocontext1);
			threadId <- liftIO (forkIO (ioProcess de ps' newIOSt >> return ()));	-- Spawn the interactive process (it blocks immediately on channel)
			liftIO (writeDeviceEvent de (ProcessDevice,ProcessInitialise));	-- Make sure the interactive process gets started
		}
	  }


{-	An ioProcess handles events coming in on its (Chan DeviceEvent) until termination (closeProcess).
-}
ioProcess :: DeviceEvents -> ps -> IOSt IF_MVAR(,ps) -> IO (ps, IOSt IF_MVAR(,ps))
ioProcess de ps ioState
	| ioStClosed ioState = return (ps, ioState)
	| otherwise = do
		reregisterProcessEventHandler ioState
		(device,deviceEvent) <- readDeviceEvent de
		(ps1, ioState1) <- handleEventForProcess device deviceEvent ps ioState
		freeDeviceEvent de
		ioProcess de ps1 ioState1
	where
#if MVAR
		handleEventForProcess :: Device -> DeviceEvent -> IOSt -> IO IOSt
		handleEventForProcess device schedulerEvent ioState
			= do {
				(_,ioState2)   <- fromGUI initIO ioState1;
				let (closed,ioState3) = ioStClosed ioState2
				in  if   closed
				    then return ioState3
				    else let (deviceFunctions,ioState4) = ioStGetDeviceFunctions ioState3
				             (found,df)                 = cselect (\df -> dDevice df == device) undefined deviceFunctions
				         in  
				         if   not found
				         then schedulerFatalError "ioProcess" "could not find Device to handle DeviceEvent"
				         else fromGUI (dDoIO df schedulerEvent) ioState4 >>= (\(_,ioState5) -> return ioState5)
			  }
			where
				(initIO,ioState1) = ioStGetInitIO ioState
#else
		handleEventForProcess :: Device -> DeviceEvent -> ps -> IOSt ps -> IO (ps, IOSt ps)
		handleEventForProcess device schedulerEvent ps ioState
			= do {
				(ps1,ioState2) <- fromGUI (initIO ps) ioState1;
				if   ioStClosed ioState2
				then return (ps1, ioState2)
				else let deviceFunctions = ioStGetDeviceFunctions ioState2
				         (found,df)      = cselect (\df -> dDevice df == device) undefined deviceFunctions
				     in  
				         if   not found
				         then schedulerFatalError "ioProcess" "could not find Device to handle DeviceEvent"
				         else fromGUI (dDoIO df schedulerEvent ps1) ioState2
			  }
			where
				initIO   = ioStGetInitIO ioState
				ioState1 = ioStSetInitIO (return) ioState
#endif
		reregisterProcessEventHandler :: IOSt IF_MVAR(,ps) -> IO ()
		reregisterProcessEventHandler ioState
			| ioStClosed  ioState = return ()
			| otherwise = do
				let context = (ioStGetContext ioState)
				iocontext <- takeMVar context
				let (maybePEH,iocontext1) = ioContextRemoveProcessEventHandler (ioStGetIOId ioState) iocontext
				(case maybePEH of
					Nothing  -> schedulerFatalError "ioProcess" "could not find ProcessEventHandler in Context"
					Just peh -> putMVar context (ioContextSetProcessEventHandler (peh {pehEventFilter=processEventFilter ioState}) iocontext1))


{-	processEventFilter passes the current Device dEvent DeviceFunctions to the
	argument SchedulerEvent to determine if this event should be handled by this
	process.
-}
	
processEventFilter :: IOSt IF_MVAR(,ps) -> SchedulerEvent -> IO (Bool,Maybe DeviceEventInfo,SchedulerEvent)
processEventFilter ioState schedulerEvent
    | ioStClosed ioState = return (False,Nothing,schedulerEvent)
    | otherwise 	 = filterEventForDevices eventFunctions schedulerEvent ioState
    where        
        eventFunctions             = [(dDevice df,dEvent df) | df <- ioStGetDeviceFunctions ioState]

        filterEventForDevices :: [(Device,EventFunction IF_MVAR(,ps))] -> SchedulerEvent -> IOSt IF_MVAR(,ps)
                              -> IO (Bool,Maybe DeviceEventInfo,SchedulerEvent)
        filterEventForDevices ((device,mapDeviceEvent):mapDeviceEvents) schedulerEvent ioState
                = do {
                    (forThisDevice,okDeviceEvent,schedulerEvent1) <- mapDeviceEvent ioState schedulerEvent;
                    if   not forThisDevice
                    then filterEventForDevices mapDeviceEvents schedulerEvent1 ioState
                    else return ( forThisDevice
                                , case okDeviceEvent of
                                       Just deviceEvent -> Just (device,deviceEvent)
                                       nothing          -> Nothing
                                , schedulerEvent1
                                )
                  }
        filterEventForDevices _ schedulerEvent _
            = return (False,Nothing,schedulerEvent)


{-	Quit this interactive process.
	Quitting a process involves the following:
	- Set the RuntimeState to Closed (quitProcess is the only function that does this)
	- Close all devices
	- Remove the process from the Context
-}
quitProcess :: IF_MVAR(GUI (),ps -> GUI ps ps)
quitProcess IF_MVAR(,ps)
	= do {
		rs <- accIOEnv ioStGetRuntimeState;
		if   rsIsClosed rs
		then IF_MVAR(return (),return ps)
		else do {
		             deviceFunctions <- accIOEnv ioStGetDeviceFunctions;
#if MVAR
		             sequence [dClose df | df<-deviceFunctions];
#else
		             ps1 <- seqListM [dClose df | df<-deviceFunctions] ps;
#endif
			     osdInfo <- accIOEnv ioStGetOSDInfo;
			     liftIO (osCloseOSDInfo osdInfo);
		             appIOEnv (ioStSetRuntimeState Closed);
		             
		             -- This part is new: remove the process from the context
		             ioId              <- accIOEnv ioStGetIOId;
		             myPEH             <- ioStRemoveProcessEventHandler ioId;
		             if   isNothing myPEH	-- This condition should not occur
		             then schedulerFatalError "quitProcess" "could not remove ProcessEventHandler from Context"
		             else return IF_MVAR((),ps1)
		        }
	     }
#if MVAR
#else
	where
		seqListM :: [ps -> GUI ps ps] -> ps -> GUI ps ps
		seqListM (m:ms) ps
			= m ps >>= (\ps -> seqListM ms ps)
		seqListM _ ps
			= return ps
#endif
